#
#   Vispan - Virus and Spam analyser
#	This program analyses the mail log file entries created by the 
#	MailScanner program written by Julian Field. It is available at
#	http://www.mailscanner.info
#
#   Copyright (C) 2002  David While
#
#   $Id: Config.pm,v 1.5 2004/12/09 11:53:02 dwhile Exp $
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, David While, can be contacted by email at
#      david.while@uce.ac.uk
#   or by paper mail at
#      David While
#      School of Computing
#      University of Central England
#      Birmingham
#      B42 2SU
#      United Kingdom
#

package Config;
use FileHandle;
use strict;

my %File;

##################################################
#                                                #
# Subroutine: Value                              #
# Parameters: Required Config file value         #
# Returns: The value                             #
# Purpose: Return the stting of a config file    #
#          entry                                 #
#                                                #
##################################################
sub Value {
   my($Name) = @_;

   return $File{$Name};
}


##################################################
#                                                #
# Subroutine: ReadConfFile                       #
# Parameters: config file name                   #
# Returns: None                                  #
# Purpose: Read the config file into a hash      #
#          array.                                #
#                                                #
##################################################
#
# Read the configuration file. Doesn't allow includes yet...
#
sub ReadConfFile {
  my($filename) = @_;

  # Slurp the whole file into a big hash.
  # Complain if we see the same keyword more than once.
  my($fileh, $linecounter, $origkey, $key, $value, $ErrorsSeen, $ErrorReport);

  $fileh = new FileHandle;
  $fileh->open("<$filename")
    or MailScanner::Log::DieLog("Could not read configuration file %s, %s",
                                $filename, $!);

  $linecounter = 0;
  $ErrorsSeen  = 0;
  while(<$fileh>) {
    $linecounter++;
    chomp;
    s/#.*$//;
    s/^\s+//;
    s/\s+$//;
    next if /^$/;

    undef $origkey;
    undef $key;
    undef $value;
    /^(.*?)\s*=\s*(.*)$/;
    ($origkey,$value) = ($1,$2);


    $key = $origkey;
    $key =~ s/[^A-Za-z0-9]//g; # Leave numbers and letters only


    if ($key eq "") {
      # Invalid line
      $ErrorReport .= "Error in line $linecounter of $filename, " .
                      "line does not make sense. ";
      $ErrorsSeen = 1;
    }
    elsif (defined $File{$key}) {
      # We've seen this keyword before.
      $ErrorReport .= "Error in line $linecounter of $filename, " .
                      "setting value of ". $origkey ." twice! ";
      $ErrorsSeen = 1;
    } else {
      $File{$key} = $value;
      #print STDERR "Defining $key = $value\n";
      #$LineNos{$key} = $linecounter; # Save where the value was stored
    }
  }
  $fileh->close();

  $File{RunInterval} = 10;

  # Now determine what graphic file type to generate

  use GD::Graph::lines;

  my $graph = GD::Graph::lines->new;
  $File{GraphicFormat} = $graph->export_format;

  # If we have seen any errors, we can't trust the line numbers in
  # %LineNos, so bail out.
  #
  # Error messages are grouped and reported together, otherwise
  # The reference to 'these errors' in the last message doesn't make
  # sense.
  if ($ErrorsSeen) {
    $ErrorReport .= "Can't continue processing configuration file " .
                    "until these errors have been corrected.";
     print $ErrorReport;
     exit 3;
  }
}

1;
