#
#   Vispan - Virus and Spam analyser
#	This program analyses the mail log file entries created by the 
#	MailScanner program written by Julian Field. It is available at
#	http://www.mailscanner.info
#
#   Copyright (C) 2002  David While
#
#   $Id: Print.pm,v 1.31 2004/11/25 11:15:02 dwhile Exp $
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, David While, can be contacted by email at
#      david.while@uce.ac.uk
#   or by paper mail at
#      David While
#      School of Computing
#      University of Central England
#      Birmingham
#      B42 2SU
#      United Kingdom
#

package Print;

use POSIX qw(strftime);
use Number::Format qw(:subs :vars);

##################################################
#                                                #
# Subroutine: Start                              #
# Parameters: File handle                        #
# Returns: None                                  #
# Purpose: Create all the menus and page stuff   #
#          that doesn't change on each page      #
#                                                #
##################################################
sub Start {

   local($handle) = @_;
   my $PageTitle = Config::Value(PageTitle);
   my $StyleSheet = Config::Value(StyleSheet);
   my $Logo = Config::Value(Logo);
   my $Header = Config::Value(Header);
   my $now = strftime "%B %e %Y %T %Z", localtime(time);
   my $expire = strftime "%a, %b %d %Y %T %Z", localtime(time);
   my $line;

   # If there's a logo specified in the config file 
   # then output it otherwise don't
   if ($Logo eq "" ) {
      $Img = "";
      }
   else {
      $Img = '<IMG SRC="'.$Logo.'" alt="logo">';
      }
   # Now output the head part of the file.
   print $handle <<"(END HEAD)";
<!DOCTYPE html 
     PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<meta http-equiv="Refresh" content="600"/>
<meta http-equiv="Expires" content="$expire"/>
<meta http-equiv="cache-control" content="no-cache"/>
<title>$PageTitle</title>
<link type="text/css" rel="stylesheet" href="$StyleSheet"/>
</head>
<body>
(END HEAD)

   # Output the contents of the specified header file if it exists
   if( open(HEADER, $Header) ) {
      @Header = <HEADER>;
      close( HEADER );
      foreach $line (@Header) {
         print $handle $line;
         }
      }

   print $handle <<"(END HTML)";
<table width="100%">
<tr><td>
$Img
</td><td>
<!-- "Welcome" text -->
<h2>$PageTitle</h2>
</td></tr></table>

<table border="0" width="100%" cellspacing="0" cellpadding="1">
<tr class="HeaderNavigation">
<td align="left">
<a class="Nav" href="10min.html">10 mins</a> &nbsp;|&nbsp; 
<a class="Nav" href="hourly.html">Hourly</a> &nbsp;|&nbsp;
<a class="Nav" href="daily.html">Daily</a> &nbsp;|&nbsp;
<a class="Nav" href="monthly.html">Monthly</a> &nbsp;|&nbsp;
<a class="Nav" href="yearly.html">Yearly</a> &nbsp;&nbsp;
</td>

<td>
Generated: $now
</td>

<td align="right">
<a class="Nav" href="index.html">Summary</a> &nbsp;|&nbsp; 
<a class="Nav" href="graphs.html">Graphs</a>&nbsp;|&nbsp;
<a class="Nav" href="spam.html">Spam</a> &nbsp;|&nbsp;
<a class="Nav" href="virus.html">Viruses</a> &nbsp;|&nbsp;
<a class="Nav" href="config.html">Config</a> &nbsp;&nbsp;
</td>
</tr>
</table>
<!-- header_eof //-->

<!-- body //-->
<!-- body_text //-->
(END HTML)
}


##################################################
#                                                #
# Subroutine: PrintEnd                           #
# Parameters: File handle                        #
# Returns: None                                  #
# Purpose: Output the footer on every page.      #
#                                                #
##################################################
sub PrintEnd {
   local($handle) = @_;
   my $line;
   my $Footer = Config::Value(Footer);

   if(open(FOOTER,$Footer)) {
      @footer=<FOOTER>;
      close(FOOTER);
      foreach $line (@footer) {
         print $handle $line;
         }
      }

   print $handle "<hr/><p>";
 
   if( $Footer eq "" ) {
      print $handle <<"(END VAL)";
         <a href="http://validator.w3.org/check?uri=referer"><img
          src="http://www.w3.org/Icons/valid-xhtml10"
          alt="Valid XHTML 1.0!" height="31" width="88" /></a>
(END VAL)
      }

   print $handle <<"(END HTML)";
Produced by <a class="footer" href="http://www.while.homeunix.net/mailstats/">
Vispan V$main::VERSION</a> \&copy; 2002-2004</p>
</body>
</html>
(END HTML)
}

##################################################
#                                                #
# Subroutine: y_format                           #
# Parameters:                                    #
# Returns: None                                  #
# Purpose: Format the numbers on the Y-axes of   #
#          the graphs                            #
#                                                #
##################################################
sub y_format{
   my ($value) = @_;

   return(format_number($value));
}

##################################################
#                                                #
# Subroutine: GenGraph                           #
# Parameters: Title - title of the graph         #
#             XLabel - the label for the X axis  #
#             File - the file name               #
#             SourceData - the data to be graphed#
# Returns: None                                  #
# Purpose: For each of the interval pages,       #
#          generate the graph                    #
#                                                #
##################################################
sub GenGraph{
   my ($Title, $XLabel, $File, $Ticks, $NumValues, @SourceData) = @_;

   use GD::Graph::lines;

   my $format = Config::Value(GraphicFormat);
   my $StartValue;
   my(@axis, @series1, @series2, @series3, @series4, @series5);
   my @Legend = ('Mail', 'Rejected', 'Spam','HighSpam','Viruses');

   # Now create the data set

   if( ($NumValues != 0) and ($NumValues < scalar @SourceData)) {
      $StartValue = (scalar @SourceData) - $NumValues;
      }
   else {
      $StartValue = 0;
      }
   for( $i = $StartValue; $i < scalar @SourceData; $i++) {
      @axis[$i-$StartValue] = $SourceData[$i]->{Time};
      @series1[$i-$StartValue] = $SourceData[$i]->{Mail};
      @series2[$i-$StartValue] = $SourceData[$i]->{FailedMail};
      @series3[$i-$StartValue] = $SourceData[$i]->{Spam};
      @series4[$i-$StartValue] = $SourceData[$i]->{HighSpam};
      @series5[$i-$StartValue] = $SourceData[$i]->{Virus};
      }

   if( Config::Value(FromLeft) == 1 ) {
      @axis = reverse(@axis);
      @series1 = reverse(@series1);
      @series2 = reverse(@series2);
      @series3 = reverse(@series3);
      @series4 = reverse(@series4);
      @series5 = reverse(@series5);
      $Offset = $Ticks - $Offset - 1;
      }
   my @Data = (\@axis, \@series1, \@series2, \@series3, \@series4, \@series5);
   # Set up the graph
   my $graph = GD::Graph::lines->new(500, 300);
   $graph->set( 
      x_label           => $XLabel,
      y_label           => 'Messages',
      title             => $Title. ' analysis',
      y_min_value	=> 0,
      x_labels_vertical	=> 1,
      x_label_skip	=> $Ticks,
      y_number_format	=> \&y_format,
      line_width	=> 2,
      box_axis		=> 0
  ) or die $graph->error;

   $graph->set_legend(@Legend);
   my $HTMLDir = Config::Value(HTMLDir);
   open(IMG, ">$HTMLDir/$File.$format") or die $!;
   binmode IMG;
   print IMG $graph->plot(\@Data)->$format();
   close IMG;  
}

##################################################
#                                                #
# Subroutine: SummaryGraph                       #
# Parameters: File - the file name               #
#             SourceData - the data              #
# Returns: None                                  #
# Purpose: Generate the index page graphs        #
#                                                #
##################################################
sub SummaryGraph{
   my ($File, $Ylabel, $Title, $Data, @SourceData) = @_;

   use GD::Graph::lines;


   my $format = Config::Value(GraphicFormat);
   my(@axis, @series1);
   my($Offset) = -1;

   # Now create the data set
   for( $i = 0; $i < scalar @SourceData; $i++) {
      if( ($SourceData[$i]->{Time} =~ /.[02468]:00/)) {
         @axis[$i] = $SourceData[$i]->{Time};
         if( $Offset == -1 ) {
            $Offset = $i;
            }
         }
      else {
         @axis[$i] = "";
         }
      @series1[$i] = $SourceData[$i]->{$Data};
      }
   if( Config::Value(FromLeft) == 1 ) {
      @axis = reverse(@axis);
      @series1 = reverse(@series1);
      $Offset = 11 - $Offset;
      }

   my @Data = (\@axis, \@series1);
   my $graph = GD::Graph::lines->new(220, 150);
   $graph->set( 
      y_label           => $Ylabel,
      title             => $Title,
      y_min_value	=> 0,
      x_labels_vertical	=> 1,
      x_label_skip	=> 12,
      x_tick_offset	=> $Offset,
      line_width	=> 2,
      box_axis		=> 0
  ) or die $graph->error;

   my $HTMLDir = Config::Value(HTMLDir);
   open(IMG, ">$HTMLDir/$File.$format") or die $!;
   binmode IMG;
   print IMG $graph->plot(\@Data)->$format();
   close IMG;  
}

##################################################
#                                                #
# Subroutine: HourGraph                          #
# Parameters: File - the file name               #
#             SourceData - the data              #
# Returns: None                                  #
# Purpose: Generate the index page graphs        #
#                                                #
##################################################
sub HourGraph{
   my ($File, $Ylabel, $Title, $Data, @SourceData) = @_;

   use GD::Graph::lines;


   my $format = Config::Value(GraphicFormat);
   my(@axis, @series1);
   my($Offset) = -100;

   # Now create the data set
   for( $i = 0; $i < scalar @SourceData; $i++) {
      if ($SourceData[$i]->{Time} =~ /00:00/) {
         if ( $Offset == -100 ) {
            $Offset = $i;
            }
         }
      if( (($i - $Offset) % 24) == 0 ) {
         @axis[$i] = $SourceData[$i]->{Time};
         }
      else {
         @axis[$i] = "";
         }
 
      @series1[$i] = $SourceData[$i]->{$Data} / 6;
      }

   if( Config::Value(FromLeft) == 1 ) {
      @axis = reverse(@axis);
      @series1 = reverse(@series1);
      $Offset = 23 - $Offset;
      }
   my @Data = (\@axis, \@series1);
   my $graph = GD::Graph::lines->new(220, 150);
   $graph->set( 
      y_label           => $Ylabel,
      title             => $Title,
      y_min_value	=> 0,
      x_labels_vertical	=> 1,
      x_label_skip	=> 24,
      x_tick_offset	=> $Offset,
      line_width	=> 2,
      box_axis		=> 0
  ) or die $graph->error;

   my $HTMLDir = Config::Value(HTMLDir);
   open(IMG, ">$HTMLDir/$File.$format") or die $!;
   binmode IMG;
   print IMG $graph->plot(\@Data)->$format();
   close IMG;  
}

##################################################
#                                                #
# Subroutine: DayGraph                           #
# Parameters: File - the file name               #
#             SourceData - the data              #
# Returns: None                                  #
# Purpose: Generate the index page graphs        #
#                                                #
##################################################
sub DayGraph{
   my ($File, $Ylabel, $Title, $Data, @SourceData) = @_;

   use GD::Graph::lines;


   my $format = Config::Value(GraphicFormat);
   my(@axis, @series1);
   my($Offset) = -1;

   # Now create the data set
   for( $i = 0; $i < scalar @SourceData; $i++) {
      if ($SourceData[$i]->{Time} =~ /^0/) {
         if( $Offset == -1 ) {
            $Offset = $i;
            }
         }

      if( (($i - $Offset) % 30) == 0 ) {
         @axis[$i] = $SourceData[$i]->{Time};
         }
      else {
         @axis[$i] = "";
         }
      @series1[$i] = $SourceData[$i]->{$Data} / 144;
      }
   if( Config::Value(FromLeft) == 1 ) {
      @axis = reverse(@axis);
      @series1 = reverse(@series1);
      $Offset = 29 - $Offset;
      }
   my @Data = (\@axis, \@series1);
   my $graph = GD::Graph::lines->new(220, 150);
   $graph->set( 
      y_label           => $Ylabel,
      title             => $Title,
      y_min_value	=> 0,
      x_labels_vertical	=> 1,
      x_label_skip	=> 30,
      x_tick_offset	=> $Offset,
      line_width	=> 2,
      box_axis		=> 0
  ) or die $graph->error;

   my $HTMLDir = Config::Value(HTMLDir);
   open(IMG, ">$HTMLDir/$File.$format") or die $!;
   binmode IMG;
   print IMG $graph->plot(\@Data)->$format();
   close IMG;  
}

##################################################
#                                                #
# Subroutine: PrintInterval                      #
# Parameters: Title - the title of the page      #
#             Xlabel - the graph xlabel          #
#             File - the file to be created      #
#             Data - the data to use             #
# Returns: None                                  #
# Purpose: Generate an HTML page for the         #
#          specified interval                    #
#                                                #
##################################################
sub PrintInterval {
   local ($Title, $XLabel, $File, $Ticks, $NumValues, @Data) = @_;

   my($i, $j);
   my $format = Config::Value(GraphicFormat);

   $file = Config::Value(HTMLDir)."/$File.html";
   open($fh, ">$file");
   Start($fh);
   print $fh "<h3>Statistics for the last $Title</h3>";
   print $fh "<table width=\"100%\">";
   print $fh "<tr><td valign=\"top\">\n";
   # Print the headers
   print $fh "<table>\n";
   print $fh "<tr><th>&nbsp;</th>";
   print $fh "<th colspan=\"2\">Last Period</th>";
   print $fh "<th>&nbsp;</th>";
   print $fh "<th colspan=\"2\">Current Period</th></tr>\n";
   $i = (scalar @Data) - 1;
   if( scalar @Data > 1 ) {
      $j = (scalar @Data) - 2;
      }
   else {
      $j = $i;
      }

   printf $fh "<tr><td>Total messages handled</td>";
   printf $fh "<td>%s</td>", format_number($Data[$j]->{Mail}+$Data[$j]->{FailedMail});
   print $fh "<td>&nbsp;</td><td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{Mail}+$Data[$i]->{FailedMail});
   print $fh "<td>&nbsp;</td></tr>\n";

   print $fh "<tr><td>Msgs handled successfully</td>";
   printf $fh "<td>%s</td>", format_number($Data[$j]->{Mail});
   print $fh "<td>&nbsp;</td><td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{Mail});
   print $fh "<td>&nbsp;</td></tr>\n";

   print $fh "<tr><td>Non Spam messages</td>";
   printf $fh "<td>%s</td>", format_number($Data[$j]->{Mail}-$Data[$j]->{Spam});
   print $fh "<td>&nbsp;</td><td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{Mail}-$Data[$i]->{Spam});
   print $fh "<td>&nbsp;</td></tr>\n";

   print $fh "<tr><td>Total data handled</td>";
   printf $fh "<td colspan=\"2\">%s</td>", format_bytes($Data[$j]->{Size});
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td colspan=\"2\">%s</td>", format_bytes($Data[$i]->{Size});
   print $fh "</tr>\n";

   print $fh "<tr><td>Avg message size</td>";
   if ( $Data[$j]->{Mail} == 0 ) {
      print $fh "<td colspan=\"2\">0 bytes</td>";
      }
   else {
      printf $fh "<td colspan=\"2\">%s bytes</td>", format_bytes($Data[$j]->{Size}/$Data[$j]->{Mail});
      }
   print $fh "<td>&nbsp;</td>";
   if ( $Data[$i]->{Mail} == 0 ) {
      print $fh "<td colspan=\"2\">0 bytes</td>";
      }
   else {
      printf $fh "<td colspan=\"2\">%s bytes</td>", format_bytes($Data[$i]->{Size}/$Data[$i]->{Mail});
      }
   print $fh "</tr>\n";

   if( $Data[$i]->{Mail} > 0 ) {
      $pc = ($Data[$i]->{Virus}/$Data[$i]->{Mail}) * 100;
      $spc = ($Data[$i]->{Spam}/$Data[$i]->{Mail}) * 100;
      $fpc = ($Data[$i]->{FailedMail}/($Data[$i]->{Mail}+$Data[$i]->{FailedMail})) * 100;
      }
   else { 
      $pc = 0; $spc = 0; $fpc = 0;
      }
   if( $Data[$i]->{Spam} > 0 ) {
      $hpc = ($Data[$i]->{HighSpam}/$Data[$i]->{Spam}) * 100;
      $Avgspamscorei = $Data[$i]->{SpamScore} / $Data[$i]->{Spam};
      }
   else { $hpc = 0; $Avgspamscorei = 0;}

   if( $Data[$i]->{HighSpam} > 0 ) {
      $ihss = $Data[$i]->{HighSpamScore} / $Data[$i]->{HighSpam};
      }
   else { $ihss = 0;}

   if( $Data[$j]->{Mail} > 0 ) {
      $jpc = ($Data[$j]->{Virus}/$Data[$j]->{Mail}) * 100;
      $jspc = ($Data[$j]->{Spam}/$Data[$j]->{Mail}) * 100;
      $jfpc = ($Data[$j]->{FailedMail}/($Data[$j]->{Mail}+$Data[$j]->{FailedMail})) * 100;
      }
   else { 
      $jpc = 0; $jspc = 0; $jfpc = 0;
      }
   if( $Data[$j]->{Spam} > 0 ) {
      $jhpc = ($Data[$j]->{HighSpam}/$Data[$j]->{Spam}) * 100;
      $Avgspamscore = $Data[$j]->{SpamScore} / $Data[$j]->{Spam};
      }
   else { $jhpc = 0; $Avgspamscore=0;}

   if( $Data[$j]->{HighSpam} > 0 ) {
      $jhss = $Data[$j]->{HighSpamScore} / $Data[$j]->{HighSpam};
      }
   else { $jhss = 0;}
   

   printf $fh "<tr><td>Messages rejected</td><td>%s</td>", format_number($Data[$j]->{FailedMail});
   printf $fh "<td>%2d%%</td>", $jfpc;
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{FailedMail});
   printf $fh "<td>%2d%%</td></tr>\n", $fpc;

   printf $fh "<tr><td>Spam received</td><td>%s</td>", format_number($Data[$j]->{Spam});
   printf $fh "<td>%2d%%</td>", $jspc;
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{Spam});
   printf $fh "<td>%2d%%</td></tr>\n", $spc;

   print $fh "<tr><td>Avg spam score</td>";
   printf $fh "<td>%s</td>", format_number($Avgspamscore,2);
   print $fh "<td>&nbsp;</td><td>&nbsp;</td>";
   printf $fh "<td>%s</td></tr>\n", format_number($Avgspamscorei,2);

   printf $fh "<tr><td>High scoring spam</td><td>%s</td>", format_number($Data[$j]->{HighSpam});
   printf $fh "<td>%2d%%</td>", $jhpc;
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{HighSpam});
   printf $fh "<td>%2d%%</td></tr>\n", $hpc;

   print $fh "<tr><td>Avg high spam score</td>";
   printf $fh "<td>%s</td>", format_number($jhss,2);
   print $fh "<td>&nbsp;</td><td>&nbsp;</td>";
   printf $fh "<td>%s</td></tr>\n", format_number($ihss,2);

   printf $fh "<tr><td>Virus infected emails</td><td>%s</td>", format_number($Data[$j]->{Virus});
   printf $fh "<td>%2d%%</td>", $jpc;
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>%s</td>", format_number($Data[$i]->{Virus});
   printf $fh "<td>%2d%%</td></tr>\n", $pc;

   print $fh "</table>";
   print $fh "</td><td>";
   Print::GenGraph($Title, $XLabel, $File, $Ticks, $NumValues, @Data);
   print $fh "<img src=\"$File.$format\" alt=\"graph\"/>";
   print $fh "</td></tr></table>\n";
   %Viruslist =  %{$Data[scalar @Data - 1]->{Virlist}};
   %SASpamTrap = %{$Data[scalar @Data - 1]->{SASpamTrap}};
   %Country = %{$Data[scalar @Data - 1]->{SCountry}};
   %VCountry = %{$Data[scalar @Data - 1]->{VCountry}};
   print $fh "<table><tr>";
   if( scalar keys %SASpamTrap > 0 ) {
      print $fh "<td valign=\"top\"><h3>Spam Trap Details</h3>\n";
      PrintSpamTrap($fh, %SASpamTrap);
      print $fh "</td>";
      }
   if( scalar keys %Country > 0 and Config::Value(UseGeoIP) == 1) {
      print $fh "<td valign=\"top\"><h3>Spam Country</h3>\n";
      PrintCountry($fh, %Country);
      print $fh "</td>";
      }
   if( scalar keys %VCountry > 0 and Config::Value(UseGeoIP) == 1) {
      print $fh "<td valign=\"top\"><h3>Virus Country</h3>\n";
      PrintCountry($fh, %VCountry);
      print $fh "</td>";
      }
   if( scalar keys %Viruslist > 0 ) {
      print $fh "<td valign=\"top\"><h3>Virus Details</h3>\n";
      PrintViruses($fh, %Viruslist);
      print $fh "</td>";
      }
   print $fh "</tr></table>";
   PrintEnd($fh);
   close $fh;
}

##################################################
#                                                #
# Subroutine: PrintSummary                       #
# Parameters: GDates - dates array               #
#             GTotals - totals array             #
#             Virs - virus array                 #
#             Traps - spam traps array           #
# Returns: None                                  #
# Purpose: Generate the index summary page.      #
#                                                #
##################################################
sub PrintSummary {
   my ($GDates, $GTotals, $Virs, $Traps) = @_;

   my($i);
   my $Count = 0;


   my $format = Config::Value(GraphicFormat);
   my $file = Config::Value(HTMLDir)."/index.html";
   open($fh, ">$file");
   Start($fh);

   my ( $date, $Duration, $DailyMail);
   my ( $mqueue, $mcount, $file);

   $date = strftime "%B %e %Y %T", localtime($GDates->{StartDate});

   # Work out the duration
   $Duration = ($GDates->{LastUpdate} - $GDates->{StartDate}) / 86400;
   if ( $Duration > 0 ) {
      $DailyMail = $GTotals->{Mail} / $Duration;
      }
   else {
      $DailyMail = 0;
      }

   # Get the length of the inbound mail queue
   opendir(QDIR, "/var/spool/mqueue.in");
   $mcount += scalar(grep { /^qf/ } readdir(QDIR));
   closedir(QDIR);

   print $fh "<h3>General Mail Statistics</h3>\n";

   # First row
   print $fh "<table><tr>\n";
   printf $fh "<td>Total messages handled</td><td>%s</td>\n",format_number($GTotals->{Mail}+$GTotals->{FailedMail});
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Total data handled</td><td>";
   printf $fh " %s bytes </td>\n", format_bytes($GTotals->{Size});
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Avg message size</td>";
   if ( $GTotals->{Mail} == 0 ) {
      print $fh "<td>0 bytes</td>\n";
      }
   else {
      printf $fh "<td>%s bytes</td>\n", format_bytes($GTotals->{Size}/$GTotals->{Mail});
      }
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Avg messages per day</td>";
   printf $fh "<td>%s</td></tr>\n", format_number($DailyMail,0);

   # Second row
   printf $fh "<tr><td>Msgs handled successfully</td>";
   printf $fh "<td>%s</td>\n", format_number($GTotals->{Mail});
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>Messages rejected</td><td>%s</td>\n",format_number($GTotals->{FailedMail});
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Rejection rate</td>";
   if ( $GTotals->{Mail} == 0 ) {
      print $fh "<td>0%</td>\n";
      }
   else {
      printf $fh "<td>%.2f%%</td>\n", ($GTotals->{FailedMail} / ($GTotals->{Mail}+$GTotals->{FailedMail})) * 100;
      }
   print $fh "<td>&nbsp;</td>";
   printf $fh "<td>Blocked IPs</td><td>%d</td></tr>\n", $GTotals->{Blocked};

   # Third row
   printf $fh "<tr><td>Spam received</td><td>%s</td>\n",format_number($GTotals->{Spam});
   print $fh "<td>&nbsp;</td>";
   if ( $GTotals->{Mail} == 0 ) {
      print $fh "<td>Spam rate</td><td>0%</td>\n";
      }
   else {
      printf $fh "<td>Spam rate</td><td>%.2f%%</td>",($GTotals->{Spam}/$GTotals->{Mail})*100;
      }
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>High Scoring Spam</td>";
   printf $fh "<td>%s</td>\n", format_number($GTotals->{HighSpam});
   print $fh "<td>&nbsp;</td>";
   if( $GTotals->{Spam} == 0 ) {
      print $fh "<td>High Scoring Spam (%)</td><td>0%</td></tr>\n";
      }
   else {
      printf $fh "<td>High Scoring Spam (%%)</td><td>%.2f%%</td></tr>\n", ($GTotals->{HighSpam} / $GTotals->{Spam}) * 100;
      }

   # Fourth row
   printf $fh "<tr><td>Virus infected emails</td><td>%s</td>\n", format_number($GTotals->{Virus});
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Infection rate</td>";
   if( $GTotals->{Mail} > 0 ) {
      printf $fh "<td>%2.2f%%</td>\n",($GTotals->{Virus}/$GTotals->{Mail})*100;
      }
   else {
      print $fh "<td>0%</td>\n";
      }
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>AV Updated</td><td>$GDates->{VirUpdate}</td>";
   print $fh "<td>&nbsp;</td>";
   print $fh "<td>Mail pending</td><td>$mcount</td></tr>";
   print $fh "</table><div>&nbsp;</div>\n";

   print $fh "<table><tr>\n";

   print $fh "<td><img src=\"load.$format\" alt=\"load\"/><br/><br/>";
   print $fh "<img src=\"delay.$format\" alt=\"delay\"/></td>\n";
   print $fh "<td><img src=\"mail.$format\" alt=\"queue\"/><br/><br/>";
   print $fh "<img src=\"failed.$format\" alt=\"mail\"/></td>\n";
   print $fh "<td><img src=\"spam.$format\" alt=\"spam\"/><br/><br/>";
   print $fh "<img src=\"virus.$format\" alt=\"virus\"/></td>\n";
   print $fh "<td><img src=\"queue.$format\" alt=\"outqueue\"/><br/><br/>";
   print $fh "<img src=\"outqueue.$format\" alt=\"failed\"/></td>\n";
   print $fh "</tr><tr><td valign=\"top\">";
   # Output the virus details
   print $fh '<table class="summary"><tr><td colspan="2" class="title">Top Viruses</td></tr>';

   my @Sorted = sort {$Virs->{$a}->{Scanner} cmp $Virs->{$b}->{Scanner} ||
                  $Virs->{$b}->{Count} <=> $Virs->{$a}->{Count} } keys %$Virs;
   my $Scanner = $Virs->{@Sorted[0]}->{Scanner};
   foreach $ThisOne (@Sorted) {
      if( $Virs->{$ThisOne}->{Scanner} ne $Scanner ) {
         next;
         }
      $count = $Virs->{$ThisOne}->{Count};
      if( $Count++ >= (Config::Value(DisplayTop) )) {
         last;
         }
      printf $fh "<tr><td class=\"summ\">$ThisOne</td><td align=\"right\" class=\"info\">%s</td></tr>\n",format_number($count);
   }

   print $fh "</table></td><td>\n";

   # Output the spam trap details
   my @SASorted = sort { $Traps->{$b} <=> $Traps->{$a} ||
			$a cmp $b} keys %$Traps;

   print $fh "<table class=\"summary\"><tr><td colspan=\"2\" class=\"title\">Top Spam Traps</td></tr>\n";
   $Count = 0;
   foreach $ThisOne (@SASorted) {
      print $fh "<tr><td class=\"summ\">$ThisOne</td>";
      print $fh "<td class=\"info\">$Traps->{$ThisOne}</td></tr>\n";
      if( $Count++ >= (Config::Value(DisplayTop) -1) ) { last; }
      }
   print $fh "</table>";
   print $fh "</td>\n";
   print $fh "</tr></table>\n";
   print $fh "<div class=\"note\">Data since $date";
   printf $fh " (%.1f days)</div>\n", $Duration;
   PrintEnd($fh);
   close $fh;
}

##################################################
#                                                #
# Subroutine: PrintGraphs                        #
# Parameters: GDates - dates array               #
#             GTotals - totals array             #
#             Virs - virus array                 #
#             Traps - spam traps array           #
# Returns: None                                  #
# Purpose: Generate the index summary page.      #
#                                                #
##################################################
sub PrintGraphs {

   my $format = Config::Value(GraphicFormat);
   my $file = Config::Value(HTMLDir)."/graphs.html";
   open($fh, ">$file");
   Start($fh);

   print $fh "<h3>10 Minute Average</h3>\n";
   print $fh "<table><tr>\n";
   print $fh "<td><img src=\"load.$format\" alt=\"load\"/><br/><br/>";
   print $fh "<img src=\"delay.$format\" alt=\"delay\"/></td>\n";
   print $fh "<td><img src=\"mail.$format\" alt=\"mail\"/><br/><br/>";
   print $fh "<img src=\"failed.$format\" alt=\"failed\"/></td>\n";
   print $fh "<td><img src=\"spam.$format\" alt=\"spam\"/><br/><br/>";
   print $fh "<img src=\"virus.$format\" alt=\"virus\"/></td>\n";
   print $fh "<td><img src=\"queue.$format\" alt=\"queue\"/><br/><br/>";
   print $fh "<img src=\"outqueue.$format\" alt=\"outqueue\"/></td>\n";
   print $fh "</tr></table>\n";

   if( -e Config::Value(HTMLDir)."/hourload.$format" ) {
      print $fh "<h3>1 Hour Average</h3>\n";
      print $fh "<table><tr>\n";
      print $fh "<td><img src=\"hourload.$format\" alt=\"load\"/><br/><br/>";
      print $fh "<img src=\"hourdelay.$format\" alt=\"delay\"/></td>\n";
      print $fh "<td><img src=\"hourmesg.$format\" alt=\"mail\"/><br/><br/>";
      print $fh "<img src=\"hourfail.$format\" alt=\"failed\"/></td>\n";
      print $fh "<td><img src=\"hourspam.$format\" alt=\"spam\"/><br/><br/>";
      print $fh "<img src=\"hourvirus.$format\" alt=\"virus\"/></td>\n";
      print $fh "<td><img src=\"hourqueue.$format\" alt=\"queue\"/><br/><br/>";
      print $fh "<img src=\"houroutqueue.$format\" alt=\"outqueue\"/></td>\n";
      print $fh "</tr></table>\n";
      }

   if( -e Config::Value(HTMLDir)."/dayload.$format" ) {
      print $fh "<h3>1 Day Average</h3>\n";
      print $fh "<table><tr>\n";
      print $fh "<td><img src=\"dayload.$format\" alt=\"load\"/><br/><br/>";
      print $fh "<img src=\"daydelay.$format\" alt=\"delay\"/></td>\n";
      print $fh "<td><img src=\"daymesg.$format\" alt=\"mail\"/><br/><br/>";
      print $fh "<img src=\"dayfail.$format\" alt=\"failed\"/></td>\n";
      print $fh "<td><img src=\"dayspam.$format\" alt=\"spam\"/><br/><br/>";
      print $fh "<img src=\"dayvirus.$format\" alt=\"virus\"/></td>\n";
      print $fh "<td><img src=\"dayqueue.$format\" alt=\"queue\"/><br/><br/>";
      print $fh "<img src=\"dayoutqueue.$format\" alt=\"outqueue\"/></td>\n";
      print $fh "</tr></table>\n";
      }
   PrintEnd($fh);
   close $fh;
}

##################################################
#                                                #
# Subroutine: PrintViruses                       #
# Parameters: handle - file handle               #
#             Data - the virus data array        #
# Returns: None                                  #
# Purpose: Print the virus details on the        #
#          interval pages                        #
#                                                #
##################################################
sub PrintViruses {
   my($handle, %Data) = @_;

   my $Columns = 2;
   my $CurrCol = 0;
   # Sort the data into scanner then by number of viruses
   my @Sorted = sort {$Data{$a}->{Scanner} cmp $Data{$b}->{Scanner} ||
                  $Data{$b}->{Count} <=> $Data{$a}->{Count} } keys %Data;
   print $handle "<table><tr>";
   my $Scanner = "";
   foreach $ThisOne (@Sorted) {
      if( $Data{$ThisOne}->{Scanner} ne $Scanner ) {
         if( $Scanner ne "" ) {
            print $handle "</table></td>";
            }
         if( $CurrCol == $Columns ) {
            print $handle "</tr><tr>";
            $CurrCol = 0;
            }
         $CurrCol++;
         $Count = 0;
         $Scanner = $Data{$ThisOne}->{Scanner};

         print $handle "<td class=\"name\">".$Scanner;

         print $handle "<table>";
         }
      $count = $Data{$ThisOne}->{Count};
      if( $Count++ >= (Config::Value(DisplayTop) )) {
         next;
         }
      printf $handle "<tr><td class=\"info\">...$ThisOne</td><td align=\"right\" class=\"info\">%s</td></tr>\n",format_number($count);
   }
   print $handle "</table></td></tr></table>";
}

##################################################
#                                                #
# Subroutine: PrintSpamTrap                      #
# Parameters: handle - file handle               #
#             Data - the spamtrap data array     #
# Returns: None                                  #
# Purpose: Print the spamtrap details on the     #
#          interval pages                        #
#                                                #
##################################################
sub PrintSpamTrap {
   my($handle, %Data) = @_;

   my $Count = 0;
   # Sort the data
   my @Sorted = sort { $Data{$b} <=> $Data{$a} ||
			$a cmp $b} keys %Data;

   print $handle "<table>";
   foreach $ThisOne (@Sorted) {
      print $handle "<tr><td class=\"info\">$ThisOne</td>";
      printf $handle "<td class=\"info\">%s</td></tr>\n", format_number($Data{$ThisOne});
      if( $Count++ >= (Config::Value(DisplayTop) -1) ) { last; }
      }
   print $handle "</table>";
}

##################################################
#                                                #
# Subroutine: PrintCountry                       #
# Parameters: handle - file handle               #
#             Data - the country data array      #
# Returns: None                                  #
# Purpose: Print the country details on the      #
#          interval pages                        #
#                                                #
##################################################
sub PrintCountry {
   my($handle, %Data) = @_;

   my $Count = 0;
   my @Sorted = sort { $Data{$b} <=> $Data{$a} ||
			$a cmp $b} keys %Data;

   print $handle "<table>";
   foreach $ThisOne (@Sorted) {
      print $handle "<tr><td class=\"info\">$ThisOne</td>";
      printf $handle "<td class=\"info\">%s</td></tr>\n", format_number($Data{$ThisOne});
      if( $Count++ >= (Config::Value(DisplayTop) -1) ) { last; }
      }
   print $handle "</table>";
}

##################################################
#                                                #
# Subroutine: PrintConfig                        #
# Parameters: VirVers - the VirVers array        #
#             Updates - the Updates array        #
# Returns: None                                  #
# Purpose: Generate the configuration page.      #
#                                                #
##################################################
sub PrintConfig {
   my($VirVers, $Updates) = @_;
                                                                                
   my $HTMLDir = Config::Value(HTMLDir);
   my $file = $HTMLDir."/config.html";
   open($fh, ">$file");
   Start($fh);

   my ($Scanner);
   my (@sorted);
                                                                                
   print $fh "<h2>System Configuration</h2>";
   print $fh "<h3>Software Versions</h3>";
   print $fh "<table><tr><th>Software</th><th>Version</th><th>Updated</th></tr>";

   @sorted = sort { $a cmp $b } keys %{$VirVers};
   
   foreach $Scanner ( @sorted ) {
      if( (lc($Scanner) ne "mailscanner") and
          (lc($Scanner) ne "spamassassin") ) {
         print $fh "<tr><td>$Scanner</td><td>$VirVers->{$Scanner}</td><td>$Updates->{$Scanner}</td></tr>";
         }
      else {
         print $fh "<tr><td>$Scanner</td><td>$VirVers->{$Scanner}</td><td>&nbsp;</td></tr>";
         }
      }
   print $fh "</table>";
   PrintEnd($fh);
   close($fh);
}

##################################################
#                                                #
# Subroutine: PrintTotalViruses                  #
# Parameters: Totals - the Totals array          #
# Returns: None                                  #
# Purpose: Generate the total virus page         #
#                                                #
##################################################
sub PrintTotalViruses {
   my($Totals) = @_;

   my $HTMLDir = Config::Value(HTMLDir);
   my $file = $HTMLDir."/virus.html";
   open($handle, ">$file");
   Start($handle);
   my %Virus = %{$Totals->{Virlist}};
   my %VCtry = %{$Totals->{VCountry}};

   # Sort the data.
   my @virussorted = sort { $Virus{$a}->{Scanner} cmp $Virus{$b}->{Scanner} || 
                    $Virus{$b}->{Count} <=> $Virus{$a}->{Count} } keys %Virus;
   my @vcountrysorted = sort { $VCtry{$b} <=> $VCtry{$a} || 
                             $a cmp $b} keys %VCtry;
   my $Columns = 3;
   my $CurrCol = 0;
   print $handle "<h2>Virus Statistics by Scanner</h2>\n";
   print $handle "<table width=\"100%\"><tr>";
   my $Scanner = "";
   foreach $ThisOne (@virussorted) {
      if( $Virus{$ThisOne}->{Scanner} ne $Scanner ) {
         if( $Scanner ne "" ) {
            print $handle "</table></td>";
            }
         if( $CurrCol == $Columns ) {
            print $handle "</tr><tr>";
            $CurrCol = 0;
            }
         $CurrCol++;
         $Count = 0;
         $Scanner = $Virus{$ThisOne}->{Scanner};

         print $handle "<td class=\"name\">".$Scanner;

         if( $VirVers{lc($Scanner)} ne "" ) {
            print $handle " - " . $VirVers{lc($Scanner)};
            if ( lc($Scanner) eq "mailscanner" ) {
               print $handle " " . $StyleEnd;
               }
            else {
               if( $Updates{lc($Scanner)} ne "" ) {
                  print $handle " updated " . $Updates{lc($Scanner)};
                  }
               }
            }
         else {
            if( $Updates{lc($Scanner)} ne "" ) {
               print $handle " - last updated " . $Updates{lc($Scanner)};
               }
            if ( lc($Scanner) eq "mailscanner" ) {
               print $handle " " . $StyleEnd;
               }
            }

          print $handle "<table>";
         }
      $count = $Virus{$ThisOne}->{Count};
      if( $Count++ >= (Config::Value(DisplayTop) )) {
         next;
         }
      printf $handle "<tr><td class=\"info\">...$ThisOne</td><td align=\"right\" class=\"info\">%s</td></tr>\n",format_number($count);
   }
   print $handle "</table></td></tr></table>";

   if( Config::Value(UseGeoIP) == 1 ) {
      print $handle "<table>\n";

      print $handle "<tr><th>Virus Country Source</th><th>Count</th></tr>\n";
      $Count = 0;
      foreach $type (@vcountrysorted) {
         $CountryCount = format_number($VCtry{$type});
         print $handle "<tr><td class=\"info\">$type</td><td align=\"right\">$CountryCount</td></tr>\n";
         if( $Count++ >= (Config::Value(DisplayTop) - 1)  ) { last; }
         }
      print $handle "</table>";
      }
   PrintEnd($handle);
   close($handle);
}

##################################################
#                                                #
# Subroutine: PrintSpam                          #
# Parameters: SpamType - The spamtype array      #
#             Totals - the totals array          #
#             Levels - The levels array          #
#             IP - the IP array                  #
# Returns: None                                  #
# Purpose: Generate the overall spam page        #
#                                                #
##################################################
sub PrintSpam {

   my($SpamType, $Totals, $Levels, $IP) = @_;

   my( $TypeCount, $ThisOne, $count, $color, $name );
   my( $Style, $StyleEnd, $level );
   my $Blocked = $Totals->{Blocked};
   my %Cntry = %{$Totals->{SCountry}};
   my %SpamTraps = %{$Totals->{SASpamTrap}};
   use integer;
   my $NumPerCol = $Blocked / 2;
   no integer;
   my $Count = 0;

   my $file = Config::Value(HTMLDir)."/spam.html";
   open($fh, ">$file");
   Start($fh);

   print $fh "<h2>Spam Statistics</h2>\n";
   print $fh "<table width=\"100%\"><tr><td>";
   print $fh "<table><tr><th>Spam Trap</th><th>Count</th></tr>\n";
   foreach $type (sort { $SpamType->{$b} <=> $SpamType->{$a} ||
                             $a cmp $b} keys %{$SpamType}) {
      $TypeCount=format_number($SpamType->{$type});
      print $fh "<tr><td class = \"info\">$type</td><td align=\"right\">$TypeCount</td></tr>\n";
      if( $type eq "SpamAssassin" ) {
         $Count = 0;
         foreach $trap (sort {$SpamTraps{$b} <=> $SpamTraps{$a} ||
                             $a cmp $b} keys %SpamTraps) {
            $TypeCount = format_number($SpamTraps{$trap});
            print $fh "<tr><td class=\"info\">...$trap</td><td align=\"right\">$TypeCount</td></tr>\n";
            if( $Count++ >= (Config::Value(DisplayTop) - 1) ) { last; }
            }
         }
     }

   print $fh "</table></td><td valign=\"top\">";
   if( Config::Value(UseGeoIP) == 1 ) {
      print $fh "<table>";
      print $fh "<tr><th>Country Source</th><th>Count</th></tr>\n";
      $Count = 0;
      foreach $type (sort { $Cntry{$b} <=> $Cntry{$a} ||
                             $a cmp $b} keys %Cntry) {
         $CountryCount = format_number($Cntry{$type});
         print $fh "<tr><td class=\"info\">$type</td><td align=\"right\">$CountryCount</td></tr>\n";
         if( $Count++ >= (Config::Value(DisplayTop) - 1)  ) { last; }
         }

      print $fh "</table>";
      }
   print $fh "</td></tr></table>";

   print $fh "<h3>Temporary Spam / Virus Source Blocks";
   print $fh "- (total $Blocked)</h3><p>";
   foreach $level (sort keys %{$Levels}) {
      print $fh "Level $level - $Levels->{$level}";
      print $fh "&nbsp;;&nbsp;";
      }
   print $fh "<br/>";
   print $fh "<em class=\"spam\"> Due to Spam</em> &nbsp;\n";
   print $fh "<em class=\"virus\"> Due to Viruses</em></p>\n";
   print $fh "<table width=\"100%\" border=\"0\"><tr><td valign = \"top\">";

   print $fh "<table><tr><th>Level</th><th>Blocked IP</th><th>Name</th></tr>\n";
   $Count = 0;
   foreach $ThisOne (sort keys %{$IP}) {
      if( $IP->{$ThisOne}->{SpamBucket} > 0 or 
		$IP->{$ThisOne}->{VirusBucket} > 0) {
         if( $IP->{$ThisOne}->{State} < 0) {
            if( $IP->{$ThisOne}->{State} == -1 ) {
               $Style = "spam";
               }
            elsif ( $IP->{$ThisOne}->{State} == -3) {
               $Style = "";
               }
            else { $Style = "virus"; }
            $name = $IP->{$ThisOne}->{Name};
	    $level = $IP->{$ThisOne}->{Level};
            print $fh "<tr class=\"$Style\"><td align=\"center\">$level</td><td>";
            print $fh "<a onclick=\"target='_blank';\" href=\"http://www.dnsstuff.com/tools/ip4r.ch?ip=$ThisOne\">$ThisOne</a></td>";
            print $fh "<td class=\"info\" align=\"right\">";
            print $fh "$name</td></tr>\n";
            if( $Count++ == $NumPerCol ) {
               $Count = 0;
               print $fh "</table></td><td valign=\"top\">";
               print $fh "<table><tr><th>Level</th><th>Blocked IP</th>";
               print $fh "<th>Name</th></tr>\n";
               }
            }
         }
      }
   print $fh "</table></td></tr></table>\n";
   PrintEnd($fh);
   close($fh);
   }
1;
