#
#   Vispan - Virus and Spam analyser
#	This program analyses the mail log file entries created by the 
#	MailScanner program written by Julian Field. It is available at
#	http://www.mailscanner.info
#
#   Copyright (C) 2002  David While
#
#   $Id: Sendmail.pm,v 1.3 2004/09/27 16:04:24 dwhile Exp $
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, David While, can be contacted by email at
#      david.while@uce.ac.uk
#   or by paper mail at
#      David While
#      School of Computing
#      University of Central England
#      Birmingham
#      B42 2SU
#      United Kingdom
#

package Sendmail;

my $AvgDelay = 0;
my $DelCount = 0;

##################################################
#                                                #
# Subroutine: ProcessLine                        #
# Parameters: Line                               #
# Returns: None                                  #
# Purpose: Process a sendmail log file line.     #
#                                                #
##################################################
sub ProcessLine {
   my($Line) = @_;

   my( $Size, $Day, $Hour, $Min, $Sec );

   SWITCH: {

      # We have rejected the mail for some reason
      ( $Line =~ /reject=/ ) && do {
         Stats::AddReject;
         last SWITCH;
         };

      # Work out the average message delay time
      ( $Line =~ /stat=Sent/ ) && do {
         ($Delay) = ($Line =~ /delay=(.*), /);
         $Day = 0;
         if( $Delay =~ /^.\+/ ) {
            ($Day,$Hour,$Min,$Sec) = ($Delay =~ /(.)\+(..):(..):(..)/);
            }
         else {
            ($Hour,$Min,$Sec) = ($Delay =~ /(..):(..):(..)/);
            }

         # Now calculate the delay time in seconds
         $AvgDelay += $Sec + ($Min*60) + ($Hour*60*60) + ($Day*60*60*24);
         Stats::AddDelay($AvgDelay);
         last SWITCH;
         };
      }
   }

1;
