#
#   Vispan - Virus and Spam analyser
#	This program analyses the mail log file entries created by the 
#	MailScanner program written by Julian Field. It is available at
#	http://www.mailscanner.info
#
#   Copyright (C) 2002  David While
#
#   $Id: Setup.pm,v 1.7 2004/09/27 16:04:24 dwhile Exp $
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, David While, can be contacted by email at
#      david.while@uce.ac.uk
#   or by paper mail at
#      David While
#      School of Computing
#      University of Central England
#      Birmingham
#      B42 2SU
#      United Kingdom
#

package Setup;

use strict;

##################################################
#                                                #
# Subroutine: BuildWorkingEnvironment            #
# Parameters: $WorkDir                           #
# Returns: None                                  #
# Purpose: Sets up the working directory where   #
#          the stats file is stored              #
#                                                #
##################################################
sub BuildWorkingEnvironment {
	my $WorkDir = shift || '/var/cache/vispan/';
	my @ParsedWorkDir = split(/\//, $WorkDir);
 	my $temp = "";
	foreach (@ParsedWorkDir) {
		$temp .= "\/$_";
		if(! -e $temp){
			mkdir($temp);
		}
	}
	#Add additional directories here
}

##################################################
#                                                #
# Subroutine: BuildHTMLEnvironment               #
# Parameters: $HTMLDir                           #
# Returns: None                                  #
# Purpose: Set up the HTML dir to put the        #
#          generated files in.                   #
#                                                #
##################################################
sub BuildHTMLEnvironment {
	my $HTMLDir = shift || '/var/www/html/mrtg/';
	my @ParsedHTMLDir = split(/\//, $HTMLDir);
 	my $temp = "";
	foreach (@ParsedHTMLDir) {
	   $temp .= "\/$_";
	   if(! -e $temp){
	      mkdir($temp);
	      }
	}
	#Add additional directories here
}

##################################################
#                                                #
# Subroutine: UpgradeConfigFile                  #
# Parameters: config file                        #
# Returns: None                                  #
# Purpose: Upgrade the config file to create any #
#          new entries and remove any unwanted   #
#          ones.                                 #
#                                                #
##################################################
# This code is derived heavily from Julian Field's 
# upgrade_mailscanner_conf script.

sub UpgradeConfigFile {
   my ($oldfname) = @_;

   use FileHandle;
   my $newfname = "Vispan.conf";

# Read in the old file to get all their current settings
my $oldfh = new FileHandle;
$oldfh->open($oldfname)
  or die "Cannot read old conf file $oldfname, $!";
my($key, $value, $origkey, $origline, %oldsettings, $ReadOldValue, %oldkeys);
my(%oldcomments, $comments);
$ReadOldValue = 0;
$comments = "";
while(<$oldfh>) {
  chomp;
  $origline = $_;
  s/#.*$//;
  s/^\s+//;
  s/\s+$//;
  ($comments .= "$origline\n"),next if /^$/;

  undef $origkey;
  undef $key;
  undef $value;
  /^(.*?)\s*=/; # \s*(.*)$/;
  $origkey = $1;
  $origline =~ /=\s*(.*)$/;
  $value = $1;

  $key = lc($origkey);
  $key =~ s/[^a-z0-9]//g; # Leave numbers and letters only

  $oldsettings{$key} = $value;
  $oldkeys{$key} = $origkey;
  $oldcomments{$key} = $comments;
  $comments = "";
  $ReadOldValue++;
}
$oldfh->close();

# create a temp file
my $tmpfh = new FileHandle "> /tmp/Vispan$$";

# Read in the new file to get all the default settings and new key names
my $newfh = new FileHandle;
$newfh->open($newfname)
  or die "Cannot read new default conf file $newfname, $!";
my($defaultvalue, $UsedOldValue, $UsedDefaultValue);
$UsedOldValue = 0;
$UsedDefaultValue = 0;
$comments = "";
while(<$newfh>) {
  chomp;
  $origline = $_;
  s/#.*$//;
  s/^\s+//;
  s/\s+$//;
  ($comments .= "$origline\n"),next if /^$/;

  undef $origkey;
  undef $key;
  undef $defaultvalue;
  /^(.*?)\s*=/; # \s*(.*)$/;
  $origkey = $1;
  /=\s*(.*)$/;
  $defaultvalue = $1;

  $key = lc($origkey);
  $key =~ s/[^a-z0-9]//g; # Leave numbers and letters only

  if (exists $oldsettings{$key}) {
    # They previously had a setting for this parameter
    print $tmpfh $oldcomments{$key};
    print $tmpfh "$origkey = $oldsettings{$key}\n";
    delete $oldsettings{$key};
    $comments = "";
    $UsedOldValue++;
  } else {
    # they are using the new default value for this parameter
    print $tmpfh $comments;
    print $tmpfh "$origline\n";
    $comments = "";
    $UsedDefaultValue++;
  }
}
$newfh->close();

# Now copy the temp file over the old config file.

my $result = `cp -f /tmp/Vispan$$ /etc/Vispan.conf`;
$result = `rm -f /tmp/Vispan$$`;
print STDERR <<EOL;

Summary
-------
Read $ReadOldValue settings from old $oldfname
Used $UsedOldValue settings from old $oldfname
Used $UsedDefaultValue default settings from new $newfname

EOL
}

1;
