#
#   Vispan - Virus and Spam analyser
#	This program analyses the mail log file entries created by the 
#	MailScanner program written by Julian Field. It is available at
#	http://www.mailscanner.info
#
#   Copyright (C) 2002  David While
#
#   $Id: Stats.pm,v 1.52 2005/01/02 11:46:18 dwhile Exp $
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#   The author, David While, can be contacted by email at
#      david.while@uce.ac.uk
#   or by paper mail at
#      David While
#      School of Computing
#      University of Central England
#      Birmingham
#      B42 2SU
#      United Kingdom
#

package Stats;

use POSIX qw(strftime);
use Number::Format qw(:subs :vars);
use Net::DNS;
use Net::CIDR;
use Storable;
#use strict;

my %Dates = (
      StartDate => 0,
      LastDate => 0,
      LastUpdate => 0,
      FilePosition => 0,
      VirUpdate => "",
      );

my %Totals;

my @Interval;
my @Hourly;
my @Daily;
my @Monthly;
my @Yearly;
my %ThisInt = (
      Virus => 0,
      Spam => 0,
      SpamScore => 0,
      HighSpam => 0,
      HighSpamScore => 0,
      FailedMail => 0,
      Mail => 0,
      AvgDelay => 0,
      DelCount => 0,
      Time => 0,
      Load => 0,
      Size => 0,
      Queue => 0,
      OutQueue => 0,
      Virlist => {},
      SASpamTrap => {},
      VCountry => {},
      SCountry => {},
      );

my %IP = {
      Spam => 0,
      Color => 0,
      State => 0,
      Name => "",
      SpamBucket => 0,
      VirusBucket => 0,
      Virus => 0,
      Grace => 0,
      Level => 0,
      };

my @Store;

my %VirusInt;
my %SASpamInt;
my %Country;
my %VCountry;

my %Updates;
my %VirVers;
my %SpamType;
my %Levels;


sub new {
   return bless{};
   }

##################################################
#                                                #
# Subroutine: ReadWorkFile                       #
# Parameters: Directory                          #
# Returns: None                                  #
# Purpose: Read in the data from the stored      #
#          file.                                 #
#                                                #
##################################################
sub ReadWorkFile {

   my ($WorkDir) = @_;
   my $SpamDecrement = 1;
   my $VirusDecrement = 1;
   my $GraceDecrement = 1;
   my ($Line, $Vir, $Count, $Domain, $Total, $SBucket, $VBucket, $Name); 
   my ($Type, $Cnty, $Date, $LogDate, $currpos, $rec, $Scanner, $VirVer);


   if( -e $WorkDir.'/vispan.dat' ) {
      @Store = @{retrieve($WorkDir.'/vispan.dat')};
      @Interval = @{@Store[0]};
      @Hourly = @{@Store[1]};
      @Daily = @{@Store[2]};
      @Monthly = @{@Store[3]};
      @Yearly = @{@Store[4]};
      %IP = %{@Store[5]};
      %Totals = %{@Store[6]};
      %Updates = %{@Store[7]};
      %VirVers = %{@Store[8]};
      %SpamType = %{@Store[9]};
      %Dates = %{@Store[10]};
      }

# Open the working file and read in the details.
   elsif (open (VIRS, "<$WorkDir/$WorkFile")) {
      $Dates{StartDate} = <VIRS>; chomp($Dates{StartDate}); 
      $Line = <VIRS>; chomp($Line);
      ($Dates{LastUpdate}, $Dates{FilePosition}) = split(/\|/, $Line);
      $Line = <VIRS>; chomp($Line);
      ($Totals{Mail}, $Totals{FailedMail},$Totals{Spam},$Totals{HighSpam}) = split(/\|/, $Line);
      $Line = <VIRS>; chomp($Line);
      ($Totals{Size},$Dates{VirUpdate},$Totals{Virus}) = split(/\|/, $Line);
      LINE: while( $Line = <VIRS> ) {
         last LINE if $Line =~ /SPAM/;
         chomp($Line);
         $rec = {};
         ($Vir, $Count, $Scanner) = split(/\|/, $Line);
         $rec->{Count} = $Count;
         $rec->{Scanner} = $Scanner;
         $Virus{$Vir} = $rec;
         }

      # Now read the spam details
      LINE: while ( $Line = <VIRS> ) {
         last LINE if $Line =~ /TYPES/;
         chomp($Line);
         $rec = {};
         $SBucket = 0; $VBucket = 0;
         ($Domain,$Count,$Total,$SBucket,$Name,$VBucket,$Grace,$Level) = split(/\|/, $Line);
         $rec->{Spam} = $Count;
         $rec->{Color} = "#000000";
         $rec->{State} = $Total;
         $rec->{Name} = $Name;
	 $rec->{Grace} = $Grace;
	 $rec->{Level} = $Level;
         if( $rec->{State} < 0 ) {
            $Totals{Blocked}++;
            }
         $rec->{SpamBucket} = $SBucket;
         $rec->{VirusBucket} = $VBucket;
         if( Config::Value(UseHeuristics) == 1 ) {
            if( $rec->{SpamBucket} > 0 ) {
               $rec->{SpamBucket} -= $SpamDecrement;
               }
            if( $rec->{VirusBucket} > 0 ) {
               $rec->{VirusBucket} -= $VirusDecrement;
               }
            if( $rec->{Grace} > 0 ) {
               $rec->{Grace} -= $GraceDecrement;
               }
            }
         $IP{$Domain} = $rec;
         }

      LINE: while ( $Line = <VIRS> ) {
         last LINE if $Line =~ /COUNTRY/;
         chomp($Line);
         ($Type,$Count) = split(/\|/, $Line);
         $SpamType{$Type} = $Count;
         }

      COUNTRY: while ( $Line = <VIRS> ) {
         last COUNTRY if $Line =~ /VCOUNTRY/;
         chomp($Line);
         ($Cnty,$Count) = split(/\|/, $Line);
         $Country{$Cnty} = $Count;
         }

      VCOUNTRY: while ( $Line = <VIRS> ) {
         last VCOUNTRY if $Line =~ /TRAPTYPE/;
         chomp($Line);
         ($Cnty,$Count) = split(/\|/, $Line);
         $VCountry{$Cnty} = $Count;
         }

      TYPE: while ( $Line = <VIRS> ) {
         last TYPE if $Line =~ /UPDATES/;
         chomp($Line);
         ($Type,$Count) = split(/\|/, $Line);
         }

      while ($Line = <VIRS> ) {
         chomp($Line);
         ($Scanner,$Date,$VirVer) = split(/\|/, $Line);
         $Updates{lc($Scanner)} = $Date;
	 $VirVers{lc($Scanner)} = $VirVer;
        }
      }
   else {
      # If no workfile exists set the start time to now.
      $Dates{StartDate} = time();
      $Dates{LastUpdate} = 0;
      my $file = Config::Value(LogFile);
      open(LOG, $file);
      seek(LOG, 0, 2);
      $Dates{FilePosition} = tell(LOG);
      close(LOG);
      Mailscanner::GetLocalVerNums();
      }
   close (VIRS);


   DoHeuristics();
}

##################################################
#                                                #
# Subroutine: WriteWorkFile                      #
# Parameters: directory                          #
# Returns: None                                  #
# Purpose: Write out the data to the storage     #
#          file.                                 #
#                                                #
##################################################
sub WriteWorkFile {

   my($WorkDir) = @_;

   $Store[0] = \@Interval;
   $Store[1] = \@Hourly;
   $Store[2] = \@Daily;
   $Store[3] = \@Monthly;
   $Store[4] = \@Yearly;
   $Store[5] = \%IP;
   $Store[6] = \%Totals;
   $Store[7] = \%Updates;
   $Store[8] = \%VirVers;
   $Store[9] = \%SpamType;
   $Store[10] = \%Dates;
   store(\@Store, $WorkDir."/vispan.dat");

   # Now store the iptables settings in case
   my $IPFile = Config::Value(IPTablesConf);
   my $IPTables = Config::Value(IPTables);
   if( Config::Value(UseIPTables) == 1 ) {
      system ("$IPTables-save > $IPFile");
      }
}

##################################################
#                                                #
# Subroutine: AddMessage                         #
# Parameters:                                    #
# Returns:                                       #
# Purpose: add one to the message count and add  #
#          the message size to the total         #
#                                                #
##################################################
sub AddMessage {
   my ($Message, $Size) = @_;

   $ThisInt{Mail} += $Message;
   if( $Size > 0 ) {
      $ThisInt{Size} += $Size;
      }
}

##################################################
#                                                #
# Subroutine: AddReject                          #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add one to the reject count           #
#                                                #
##################################################
sub AddReject {
   $ThisInt{FailedMail}++;
}

##################################################
#                                                #
# Subroutine: AddSpamTrap                        #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add the spam trap to the list         #
#                                                #
##################################################
sub AddSpamTrap {
   my ($Trap) = @_;

   $SpamType{$Trap}++;
}

##################################################
#                                                #
# Subroutine: AddSpamAssassinTrap                #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add the SpamAssasin trap to the list  #
#                                                #
##################################################
sub AddSpamAssassinTrap {
   my ($Trap) =@_;

   $SASpamInt{$Trap}++;
}

##################################################
#                                                #
# Subroutine: AddVirDate                         #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add the update time for the specified #
#          virus scanner                         #
#                                                #
##################################################
sub AddVirDate {
   my ($Date, $Scanner) = @_;

   $Updates{lc($Scanner)} = $Date;
   if( (lc($Scanner) ne 'mailscanner') and
       (lc($Scanner) ne 'spamassassin') ) {
      $Dates{VirUpdate} = $Date;
      }
}

##################################################
#                                                #
# Subroutine: AddVirVersion                      #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add the version number for the        #
#          specified scanner                     #
#                                                #
##################################################
sub AddVirVersion {
   my ($Scanner, $VirVer) = @_;
   my $VerNum;

      $VirVers{lc($Scanner)} = $VirVer;
}

##################################################
#                                                #
# Subroutine: AddDelay                           #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add to the delay count and add the    #
#          actual delay                          #
#                                                #
##################################################
sub AddDelay {
   my ($Delay) = @_;

   $ThisInt{AvgDelay} = $Delay;
   $ThisInt{DelCount}++;
}

##################################################
#                                                #
# Subroutine: AddVirus                           #
# Parameters:                                    #
# Returns:                                       #
# Purpose: add the virus detected to the         #
#          appropriate scanners list             #
#                                                #
##################################################
sub AddVirus {
   my($Vir,$Scanner) =@_;
   $VirusInt{$Vir}->{Count}++;
   $VirusInt{$Vir}->{Scanner} = $Scanner;
}

##################################################
#                                                #
# Subroutine: FoundDomain                        #
# Parameters: Addr - the IP address              #
#             Score - the spam score or 0 for    #
#                     viruses                    #
#             Type - spam or virus               #
# Returns:                                       #
# Purpose: Run the heuristics on the IP address  #
#          to determine what action to take      #
#                                                #
##################################################
sub FoundDomain {
   my($Addr, $Score, $Type) = @_;
   my $Cntry;
   my $Threshold = (40 * 60) / Config::Value(RunInterval);
   use integer;
   my $VirusIncrement = ($Threshold+10) / Config::Value(NumViruses);
   my $SpamIncrement = ($Threshold+10) / Config::Value(NumSpams);
   no integer;
   if( $Addr ne "") {
      # Add one to the score 
      if( $Type eq 'S' ) {
         $IP{$Addr}->{Spam}++;
         }
      else {
         $IP{$Addr}->{Virus}++;
         }
      if( (Config::Value(UseHeuristics) == 1) ) {
         # If its not blocked then add the increment 
         # to the appropriate bucket. If it's already blocked
         # then we have a problem since no mail should be getting
         # through!
         if( $IP{$Addr}->{State} == 0 ) {
            if( $Type eq 'S' ) {
               $IP{$Addr}->{SpamBucket} += $SpamIncrement;
               }
            else {
               $IP{$Addr}->{VirusBucket} += $VirusIncrement;
               }
	    if( ($IP{$Addr}->{VirusBucket} >= $Threshold)
                   or ($IP{$Addr}->{Virus} > Config::Value(NumViruses))) {
               AddDomain($Addr, 'V');
               $IP{$Addr}->{Color} = "#FFFF00";
               }
            elsif( ($IP{$Addr}->{SpamBucket} >= $Threshold) 
                   or ($IP{$Addr}->{Spam} > Config::Value(NumSpams))) {
               AddDomain($Addr, 'S');
               $IP{$Addr}->{Color} = "#ff0000";
               }
            }
         }
      else {
         $IP{$Addr}->{SpamBucket} = 999999;
         $IP{$Addr}->{VirusBucket} = 999999;
         }

      if( Config::Value(UseGeoIP) == 1 ) {
         require Geo::IP;
         my $gi = Geo::IP->new(GEIP_STANDARD);
         $Cntry = $gi->country_name_by_addr($Addr);
         if( $Cntry eq "" ) {
            $Cntry = "Unknown";
            }
         }
      if($Type eq 'S') {
         if( $Cntry ne "" ) {
            $Country{$Cntry}++;
            }
         $ThisInt{Spam}++;
         $ThisInt{SpamScore} += $Score;
         }
      else {
         if( $Cntry ne "" ) {
            $VCountry{$Cntry}++;
            }
         $ThisInt{Virus}++;
         }
      if( $Score > Config::Value(HighScore) ) {
         $ThisInt{HighSpam}++;
         $ThisInt{HighSpamScore} += $Score;
         }
      }
   else {
      print "No IP Address\n";
      }
   }

##################################################
#                                                #
# Subroutine: AddDomain                          #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Add the IP address to the access      #
#          block file and set the counters       #
#                                                #
##################################################
sub AddDomain {
   my($BadDomain, $type) = @_;
   my ($NoBlock) = 0;
   my ($GoodIP) = 0;
   my $res = Net::DNS::Resolver->new;
   my ($AccessList, $MakeMap,@WhiteList);
   my ($BlockTime);
   my $AccessNotify = Config::Value(AccessNotify);
   my $IPTables = Config::Value(IPTables);

   $IP{$BadDomain}->{Name} = "No reverse DNS";
   my $query = $res->search($BadDomain);
   if ( $query) {
      foreach my $rr ($query->answer) {
         next unless $rr->type eq "PTR";
         $IP{$BadDomain}->{Name} = $rr->ptrdname;
         }
      }

   # Don't block if it is in the whitelist.
   @WhiteList = split(/ /,Config::Value(WhiteList));
   foreach $entry (@WhiteList) {
      @cidrlist = Net::CIDR::cidradd($entry,@cidrlist);
      }
   if( Config::Value(UseAccess) == 1 or 
	Config::Value(UseIPTables) == 1 ) {
      $AccessList = Config::Value(AccessList);
      $NoBlock = Net::CIDR::cidrlookup($BadDomain, @cidrlist);

      if( Config::Value(UseAccess) == 1 and 
		Config::Value(UseIPTables) == 0 )  {
         # Check to see if the IP address is already in the file
         # If so don't add it.
         if( $NoBlock == 0 ) {
            open FILE, "<$AccessList" or die "Can't open $AccessList: $!";
            $NoBlock = grep /$BadDomain/i, <FILE>;
            close FILE;
            }

         # Edit the file and add the appropriate entry
         if ( $NoBlock == 0 ) {
            $MakeMap = Config::Value(Makemap);
            open(ACC,">>$AccessList");
            if( $type eq 'S') {
               print ACC "$BadDomain\t550 ".Config::Value(SpamRejectText)."\n";
               }
            else {
               print ACC "$BadDomain\t550 ".Config::Value(VirusRejectText)."\n";
               }
            close ACC; 
            # Now make the database using the makemap command
            system("$MakeMap hash $AccessList < $AccessList");
            }
         }

      if( Config::Value(UseIPTables) == 1  and $NoBlock == 0 ) {
         system("$IPTables -A INPUT -s $BadDomain -p tcp --destination-port smtp -j DROP");
         }
         

      if( $NoBlock == 0 ) {
         if( $type eq 'S') {
            $IP{$BadDomain}->{State} = -1;
            }
         else {
            $IP{$BadDomain}->{State} = -2;
            }
         $Totals{Blocked}++;

         # Set the block time
         $BlockTime = Config::Value(BlockTime) / 10;
         $BlockTime = $BlockTime * (2 ** $IP{$BadDomain}->{Level});
         $IP{$BadDomain}->{Grace} = 0;

         if( $type eq 'S' ) {
            $IP{$BadDomain}->{SpamBucket} = $BlockTime;
            }
         else {
            $IP{$BadDomain}->{VirusBucket} = $BlockTime;
            }
         $IP{$BadDomain}->{Level}++;
         $Levels{$IP{$BadDomain}->{Level}}++;

         # Added for Access notification - Start
         if ($AccessNotify == 1) {
            AccessUpdateNotification($BadDomain, $IP{$BadDomain}->{Name}, $IP{$BadDomain}->{Level}, $type);
            }
         # Added for Access notification - End
         }

      else {
         $IP{$BadDomain}->{State} = -3;
         }
      }
}

##################################################
#                                                #
# Subroutine: RemoveDomain                       #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Remove the access block on the        #
#          specified IP address                  #
#                                                #
##################################################
sub RemoveDomain {
   local ($BadDomain) = @_;
   my @AllEntries;
   my ($AccessList) = Config::Value(AccessList);
   my $MakeMap = Config::Value(MakeMap);
   my $IPTables = Config::Value(IPTables);

   if( $IP{$BadDomain}->{State} != -3 ) {
      if( Config::Value(UseIPTables) == 1 ) {
         system("$IPTables -D INPUT -s $BadDomain -p tcp --destination-port smtp -j DROP &> /dev/null");
         }
      if ( Config::Value(UseAccess) == 1 ) {
         open (ACC, "+<$AccessList");
         @AllEntries = <ACC>;
         truncate(ACC, 0);
         seek(ACC,0,0);
         foreach $Entry (@AllEntries) {
            if ($Entry =~ /$BadDomain/ ) {
               # don't want to write this entry
               }	
            else {
               print ACC $Entry;
               }
            }	
         close(ACC);
         system("$Makemap hash $AccessList < $AccessList");
         }
      }

   # Now set the Grace period so that we can escalate if necessary.

   $IP{$BadDomain}->{Grace} = 1008; # 7 days
   $IP{$BadDomain}->{State} = 0;
}

##################################################
#                                                #
# Subroutine: CommandRemove                      #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Remove the access block on the        #
#          specified IP address  called from     #
#          the command line                      #
##################################################
sub CommandRemove {
   my ($IPAddr) = @_;

   if( $IP{$IPAddr}->{State} < 0 ) {
      RemoveDomain($IPAddr);
      return(1);
      }
   else {
      print "IP address $IPAddr is not blocked\n";
      return(0);
      }
}

sub mailq_dir {
   my ($mqueue) = @_;
   my @rv;
   if (!$mqueue) { @rv = ( "/var/spool/mqueue.in" ); }
   elsif ($mqueue =~ /\*|\?/) {
      @rv = split(/\s+/, `echo $mqueue`);
      }
   else {
      @rv = ( $mqueue );
      }
   push(@rv, split(/\s+/, Config::Value(QueueDirs)));
   return @rv;
   }

##################################################
#                                                #
# Subroutine: SetDate                            #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Set the date                          #
#                                                #
##################################################
sub SetDate {

   my($Field, $Value) = @_;

   $Dates{$Field} = $Value;
}

##################################################
#                                                #
# Subroutine: GetDate                            #
# Parameters:                                    #
# Returns:                                       #
# Purpose: get a date field                      #
#                                                #
##################################################
sub GetDate {

   my($Field) = @_;

   return $Dates{$Field};
}

##################################################
#                                                #
# Subroutine: GetInterval                        #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Get a value from the Interval array   #
#                                                #
##################################################
sub GetInterval {
   my ($Field) = @_;

   return $ThisInt{$Field};
}

##################################################
#                                                #
# Subroutine: AccessUpdateNotification           #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Send an email when an IP has been     #
#          added to the access block             #
#                                                #
##################################################
sub AccessUpdateNotification {

   my($IP, $name, $level, $type) = @_;

   use Mail::Sendmail 0.75;

   if( Config::Value(UseAccess) == 1 ) {
      $mail{Smtp} = Config::Value(SMTPServer);
      $mail{From} = Config::Value(NotifyFromAddress);
      $mail{To} = Config::Value(NotifyToAddress);
      $mail{Subject} = Config::Value(NotifySubject);
      $AccessList = Config::Value(AccessList);

      $mail{Message} = "\nThe following address has been blocked
\n\n   " . $IP . "\t" . $name;
      if($type eq 'S') {
         $mail{Message} .= "\nIt has been added at level $level for sending Spam\n";
         }
      else {
         $mail{Message} .= "\nIt has been added at level $level for sending Viruses\n";
         }

      sendmail(%mail) || print "Error: ($Mail::Sendmail::error)\nLog:
$Mail::Sendmail::log\n\n";
   }
}

##################################################
#                                                #
# Subroutine: CombineStats                       #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Roll the interval stats into the      #
#          other stats                           #
#                                                #
##################################################
sub CombineStats {
   my ($Gap, $Period) = @_;
   my $k, $v;
   local %Combined;
   local %Spam;
   local %Virs;
   local %Ctry;
   local %VCtry;

   $Combined{Time} = $Period->{Time};
   $Combined{Blocked} = $Period->{Blocked};
   $Combined{Load} = $Period->{Load} + $Gap->{Load};
   $Combined{Mail} = $Period->{Mail} + $Gap->{Mail};
   $Combined{Spam} = $Period->{Spam} + $Gap->{Spam};
   $Combined{SpamScore} = $Period->{SpamScore} + $Gap->{SpamScore};
   $Combined{HighSpam} = $Period->{HighSpam} + $Gap->{HighSpam};
   $Combined{HighSpamScore} = $Period->{HighSpamScore} + $Gap->{HighSpamScore};
   $Combined{FailedMail} = $Period->{FailedMail} + $Gap->{FailedMail};
   $Combined{Virus} = $Period->{Virus} + $Gap->{Virus};
   $Combined{Size} = $Period->{Size} + $Gap->{Size};
   $Combined{Queue} = $Period->{Queue} + $Gap->{Queue};
   $Combined{OutQueue} = $Period->{OutQueue} + $Gap->{OutQueue};
   $Combined{AvgDelay} = $Period->{AvgDelay} + $Gap->{AvgDelay};
   foreach $Trapref ( \%{$Gap->{SASpamTrap}}, \%{$Period->{SASpamTrap}} ) {
      while( ($k, $v) = each %$Trapref ) {
         $Spam{$k} += $v;
         }
      }
   foreach $Virref ( \%{$Gap->{Virlist}}, \%{$Period->{Virlist}} ) {
      while( ($k, $v) = each %$Virref ) {
         $Virs{$k}->{Count} += $v->{Count};
         $Virs{$k}->{Scanner} = $v->{Scanner};
         }
      }
   foreach $Countryref ( \%{$Gap->{SCountry}}, \%{$Period->{SCountry}} ) {
      while( ($k, $v) = each %$Countryref ) {
         $Ctry{$k} += $v;
         }
      }
   foreach $Countryref ( \%{$Gap->{VCountry}}, \%{$Period->{VCountry}} ) {
      while( ($k, $v) = each %$Countryref ) {
         $VCtry{$k} += $v;
         }
      }
   $Period->{Queue} = $Interval->{Queue} / scalar @Interval;
   $Combined{SASpamTrap} = \%Spam;
   $Combined{Virlist} = \%Virs;
   $Combined{SCountry} = \%Ctry;
   $Combined{VCountry} = \%VCtry;
   return %Combined;
}

##################################################
#                                                #
# Subroutine: MoveStats                          #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Roll the stats round at each time     #
#          interval. Do the necessary on the     #
#          relevant boundaries                   #
#                                                #
##################################################
sub MoveStats {

   my $min = (localtime($Dates{LastUpdate}))[1];
   my $hour = (localtime($Dates{LastUpdate}))[2];
   my $day = (localtime($Dates{LastUpdate}))[3];
   my $mon = (localtime($Dates{LastUpdate}))[4];
   my %ThisHour = {};
   my %ThisDay = {};
   my %ThisMonth = {};
   my %ThisYear = {};
   my %Viruses, %SATraps;
   my $Load = `/usr/bin/uptime`;
   my $time = strftime "%H:%M", localtime($Dates{LastUpdate});
   my $day = strftime "%d %b", localtime($Dates{LastUpdate});
   my $month = strftime "%b %y", localtime($Dates{LastUpdate});
   my $year = strftime "%Y", localtime($Dates{LastUpdate});
   my ($mqueue, $mocunt);


   foreach $mqueue (mailq_dir("/var/spool/mqueue")) {
      opendir(QDIR, $mqueue);
      $mcount += scalar(grep { /^qf/ } readdir(QDIR));
      closedir(QDIR);
      }
   $ThisInt{OutQueue} = $mcount;

   $mcount = 0;
   opendir(QDIR, "/var/spool/mqueue.in");
   $mcount += scalar(grep { /^qf/ } readdir(QDIR));
   closedir(QDIR);
   $ThisInt{Queue} = $mcount;

   ($ThisInt{Load}) = ($Load =~ /load average: .*?, (.*?),/);
   $ThisInt{Time} = $time;
   $ThisInt{Virlist} = \%VirusInt;
   $ThisInt{SASpamTrap} = \%SASpamInt;
   $ThisInt{SCountry} = \%Country;
   $ThisInt{VCountry} = \%VCountry;
   if( $ThisInt{DelCount} > 0 ) {
      $ThisInt{AvgDelay} = $ThisInt{AvgDelay} / $ThisInt{DelCount};
      }
   else {
      $ThisInt{AvgDelay} = 0;
      }
   # First Rotate the 10 minute values.
   if( (scalar @Interval) >= 144 ) {
      shift @Interval;
      }
   push(@Interval, \%ThisInt);
   Print::PrintInterval('10 Minutes', 'Minutes', '10min', 1, 12, @Interval);
   if( (scalar @Daily) > 0 ) {
      %Viruses = %{$Daily[scalar @Daily - 1]->{Virlist}};
      %SATraps = %{@Daily[scalar @Daily - 1]->{SASpamTrap}};
      }
   else { %Viruses = {}; %SATraps = {}; }

   # Now if we are on the hour update the hourly values and generate the
   # page.
   if( scalar @Hourly > 0 ) {
      %ThisHour = %{$Hourly[scalar @Hourly - 1]};
      %ThisHour = CombineStats(\%ThisInt, \%ThisHour);
      $ThisHour{Time} = $time;
      $Hourly[scalar @Hourly - 1] = \%ThisHour;
      }
   else {
      $blankref = new;
      $blankref->{Time} = $time;
      push(@Hourly, $blankref);
      }
   Print::PrintInterval('Hour', 'Hours', 'hourly', 1, 24, @Hourly);
   if( $min == 0 ) {
      Print::HourGraph('hourmesg','Messages','Messages Handled','Mail',@Hourly);
      Print::HourGraph('hourload','Load','System Load', 'Load',@Hourly);
      Print::HourGraph('hourfail','Messages','Messages Rejected',
				'FailedMail',@Hourly);
      Print::HourGraph('hourspam','Messages','Spam Handled','Spam',@Hourly);
      Print::HourGraph('hourvirus','Messages','Viruses Handled',
				'Virus',@Hourly);
      Print::HourGraph('hourqueue','Messages','Inbound Queue','Queue',@Hourly);
      Print::HourGraph('houroutqueue','Messages','Outbound Queue',
				'OutQueue',@Hourly);
      Print::HourGraph('hourdelay','Time (s)','Message Delay',
				'AvgDelay',@Hourly);
      $blankref = new;
      $blankref->{Time} = $time;
      push(@Hourly, $blankref);
      if( (scalar @Hourly) > 168 ) {
         shift @Hourly;
         }
      }

   # Now if we are at the start of the day do the daily figures
   if( scalar @Daily > 0 ) {
      %ThisDay = %{$Daily[scalar @Daily -1 ]};
      %ThisDay = CombineStats(\%ThisInt, \%ThisDay);
      $Daily[scalar @Daily - 1] = \%ThisDay;
      }
   else {
      $blankref= new;
      $blankref->{Time} = $day;
      push(@Daily, $blankref);
      }
   Print::PrintInterval('Day', 'Day', 'daily', 1, 30, @Daily);
   if( ($min == 0) and ($hour == 0) ) {
      Print::DayGraph('daymesg','Messages','Messages Handled','Mail',@Daily);
      Print::DayGraph('dayload','Load','System Load', 'Load',@Daily);
      Print::DayGraph('dayfail','Messages','Messages Rejected',
				'FailedMail',@Daily);
      Print::DayGraph('dayspam','Messages','Spam Handled','Spam',@Daily);
      Print::DayGraph('dayvirus','Messages','Viruses Handled',
				'Virus',@Daily);
      Print::DayGraph('dayqueue','Messages','Inbound Queue','Queue',@Daily);
      Print::DayGraph('dayoutqueue','Messages','Outbound Queue',
				'OutQueue',@Daily);
      Print::DayGraph('daydelay','Time (s)','Message Delay',
				'AvgDelay',@Daily);
      $blankref = new;
      $blankref->{Time} = $day;
      push(@Daily, $blankref);
      if( (scalar @Daily) > 365 ) {
         shift @Daily;
         }
      }

   # If we are at the start of a month do the monthly stats.
   if( scalar @Monthly > 0 ) {
      %ThisMonth = %{$Monthly[scalar @Monthly - 1]};
      %ThisMonth = CombineStats(\%ThisInt, \%ThisMonth);
      $Monthly[scalar @Monthly - 1] = \%ThisMonth;
      }
   else {
      $blankref = new;
      $blankref->{Time} = $month;
      push(@Monthly, $blankref);
      }
   Print::PrintInterval('Month', 'Month', 'monthly', 1, 0, @Monthly);
   if( ($min == 0) and ($hour == 0) and ($day == 1) ) {
      $blankref = new;
      $blankref->{Time} = $month;
      push(@Monthly, $blankref);
      if( (scalar @Monthly) > 12 ) {
         shift @Monthly;
         }
      }

   # If we are at the start of a year do the yearly stats.
   if( scalar @Yearly > 0 ) {
      %ThisYear = %{$Yearly[scalar @Yearly - 1]};
      %ThisYear = CombineStats(\%ThisInt, \%ThisYear);
      $Yearly[scalar @Yearly - 1] = \%ThisYear;
      }
   else {
      $blankref = new;
      $blankref->{Time} = $year;
      push(@Yearly, $blankref);
      }
   Print::PrintInterval('Year', 'Year', 'yearly', 1, 0, @Yearly);
   if( ($min == 0) and ($hour == 0) and ($day == 1) and ($mon == 0)) {
      $blankref = new;
      $blankref->{Time} = $year;
      push(@Yearly, $blankref);
      }

   # Now generate the pages.
   %Totals = CombineStats(\%ThisInt, \%Totals);
   $Totals{Blocked} = 0;
   foreach $t (keys %Levels) {
      $Totals{Blocked} += $Levels{$t};
      }
   Print::PrintGraphs();
   Print::PrintTotalViruses(\%Totals);
   Print::PrintSpam(\%SpamType, \%Totals, \%Levels, \%IP);
   Print::PrintConfig(\%VirVers, \%Updates);
   Print::PrintSummary(\%Dates, \%Totals, \%Viruses, \%SATraps);
   Print::SummaryGraph("load",'Load','System Load','Load',@Interval); 
   Print::SummaryGraph("queue",'Messages','Inbound Queue','Queue', @Interval);
   Print::SummaryGraph("outqueue",'Messages','Outbound Queue','OutQueue', @Interval);
   Print::SummaryGraph("mail",'Messages','Messages Handled','Mail', @Interval);
   Print::SummaryGraph("delay",'Time (s)','Message Delay','AvgDelay', @Interval);
   Print::SummaryGraph("spam",'Messages','Spam Handled','Spam', @Interval);
   Print::SummaryGraph("virus",'Messages','Viruses handled','Virus', @Interval);
   Print::SummaryGraph("failed",'Messages','Messages Rejected','FailedMail',@Interval);
}

##################################################
#                                                #
# Subroutine:  DoHeuristics                      #
# Parameters:                                    #
# Returns:                                       #
# Purpose: Perform the necessary actions on the  #
#          stats                                 #
#                                                #
##################################################
sub DoHeuristics{

   my $SpamDecrement = 1;
   my $VirusDecrement = 1;
   my $GraceDecrement = 1;

   $Totals{Blocked} = 0;
   foreach $rec ( keys %IP ) {
      if( Config::Value(UseHeuristics) == 1 ) {
         if( $IP{$rec}->{SpamBucket} > 0 ) {
            $IP{$rec}->{SpamBucket} -= $SpamDecrement;
            if( $IP{$rec}->{SpamBucket} == 0 and 
		$IP{$rec}->{State} == -1 ) {
               RemoveDomain($rec);
               $Totals{Blocked}--;
               }
            }
         if( $IP{$rec}->{VirusBucket} > 0 ) {
            $IP{$rec}->{VirusBucket} -= $VirusDecrement;
            if( $IP{$rec}->{VirusBucket} == 0 and 
		$IP{$rec}->{State} == -2 ) {
               RemoveDomain($rec);
               $Totals{Blocked}--;
               }
            }
         if( $IP{$rec}->{Grace} > 0 ) {
            $IP{$rec}->{Grace} -= $GraceDecrement;
            }
         }
         
      if( $IP{$rec}->{State} == -1 or $IP{$rec}->{State} == -2 ) {
         $Levels{$IP{$rec}->{Level}}++;
         $Totals{Blocked}++;
         }
      else {
         if( $IP{$rec}->{SpamBucket} == 0 and $IP{$rec}->{VirusBucket} == 0
              and $IP{$rec}->{Grace} == 0 ) {
            delete ($IP{$rec});
           }
         }
      }
}

1;
